/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

NVCV_DETAIL_DEF_TLAYOUT(W)
NVCV_DETAIL_DEF_TLAYOUT(WC)
NVCV_DETAIL_DEF_TLAYOUT(CW)

NVCV_DETAIL_DEF_TLAYOUT(NW)
NVCV_DETAIL_DEF_TLAYOUT(NWC)
NVCV_DETAIL_DEF_TLAYOUT(NCW)

NVCV_DETAIL_DEF_TLAYOUT(HW)
NVCV_DETAIL_DEF_TLAYOUT(NHW)

NVCV_DETAIL_DEF_TLAYOUT(FHW)
NVCV_DETAIL_DEF_TLAYOUT(NFHW)

NVCV_DETAIL_DEF_TLAYOUT(CHW)
NVCV_DETAIL_DEF_TLAYOUT(NCHW)
NVCV_DETAIL_DEF_TLAYOUT(HWC)
NVCV_DETAIL_DEF_TLAYOUT(NHWC)

NVCV_DETAIL_DEF_TLAYOUT(CFHW)
NVCV_DETAIL_DEF_TLAYOUT(FCHW)
NVCV_DETAIL_DEF_TLAYOUT(FHWC)
NVCV_DETAIL_DEF_TLAYOUT(NCFHW)
NVCV_DETAIL_DEF_TLAYOUT(NFCHW)
NVCV_DETAIL_DEF_TLAYOUT(NFHWC)

NVCV_DETAIL_DEF_TLAYOUT(DHW)
NVCV_DETAIL_DEF_TLAYOUT(NDHW)

NVCV_DETAIL_DEF_TLAYOUT(CDHW)
NVCV_DETAIL_DEF_TLAYOUT(DHWC)
NVCV_DETAIL_DEF_TLAYOUT(NCDHW)
NVCV_DETAIL_DEF_TLAYOUT(NDHWC)

NVCV_DETAIL_DEF_TLAYOUT(FDHW)
NVCV_DETAIL_DEF_TLAYOUT(NFDHW)

NVCV_DETAIL_DEF_TLAYOUT(CFDHW)
NVCV_DETAIL_DEF_TLAYOUT(FCDHW)
NVCV_DETAIL_DEF_TLAYOUT(FDHWC)
NVCV_DETAIL_DEF_TLAYOUT(NCFDHW)
NVCV_DETAIL_DEF_TLAYOUT(NFCDHW)
NVCV_DETAIL_DEF_TLAYOUT(NFDHWC)
