/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: Apache-2.0
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#ifndef NVCV_TEST_UNIT_CORE_DEFINITIONS_HPP
#define NVCV_TEST_UNIT_CORE_DEFINITIONS_HPP

#include <common/CheckStatus.hpp>
#include <common/Printers.hpp>
#include <gmock/gmock-matchers.h>
#include <gtest/gtest.h>

#endif // NVCV_TEST_UNIT_CORE_DEFINITIONS_HPP
