/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

document.addEventListener("DOMContentLoaded", function() {
    var params = window.location.search.substring(1).split("&").reduce(function(params, param) {
        if (!param)
        {
            return params;
        }

        var values = param.split("=");
        var name = values[0];
        var value = values[1];
        params[name] = value;
        return params;
    }, {});

    var form = document.getElementById("feedback-form");
    for (var name in params)
    {
        var input = form.querySelector("[name=" + name + "]");
        input.value = params[name];
    }
});
