/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "ColorFormat.hpp"

namespace nvcv::priv {

bool operator==(const ColorFormat &a, const ColorFormat &b)
{
    if (a.model == b.model)
    {
        if (a.model == NVCV_COLOR_MODEL_RAW)
        {
            return a.raw == b.raw;
        }
        else
        {
            return a.cspec == b.cspec;
        }
    }
    else
    {
        return false;
    }
}

bool operator!=(const ColorFormat &a, const ColorFormat &b)
{
    return !operator==(a, b);
}

} // namespace nvcv::priv
