/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef NVCV_DETAIL_INDEXSEQUENCE_HPP
#define NVCV_DETAIL_INDEXSEQUENCE_HPP

#include <cstddef>
#include <type_traits>

namespace nvcv { namespace detail {

template<std::size_t... II>
struct IndexSequence
{
};

template<std::size_t N, std::size_t... II>
struct MakeIndexSequenceImpl
{
    using type = typename MakeIndexSequenceImpl<N - 1, N, II...>::type;
};

template<std::size_t... II>
struct MakeIndexSequenceImpl<0, II...>
{
    using type = IndexSequence<(II - 1)...>;
};

template<std::size_t N>
using MakeIndexSequence = typename MakeIndexSequenceImpl<N>::type;

}} // namespace nvcv::detail

#endif // NVCV_DETAIL_INDEXSEQUENCE_HPP
